package com.egommerce.apigateway;

import java.io.File;
import java.net.InetSocketAddress;
import java.net.SocketAddress;
import java.util.HashMap;
import java.util.Map;

import javax.inject.Inject;
import javax.inject.Singleton;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.egommerce.apigateway.module.ApiGwFiltersModule;
import com.netflix.appinfo.ApplicationInfoManager;
import com.netflix.config.DynamicIntProperty;
import com.netflix.config.DynamicStringProperty;
import com.netflix.discovery.EurekaClient;
import com.netflix.netty.common.accesslog.AccessLogPublisher;
import com.netflix.netty.common.channel.config.ChannelConfig;
import com.netflix.netty.common.channel.config.ChannelConfigValue;
import com.netflix.netty.common.channel.config.CommonChannelConfigKeys;
import com.netflix.netty.common.metrics.EventLoopGroupMetrics;
import com.netflix.netty.common.proxyprotocol.StripUntrustedProxyHeadersHandler;
import com.netflix.netty.common.ssl.ServerSslConfig;
import com.netflix.netty.common.status.ServerStatusManager;
import com.netflix.spectator.api.Registry;
import com.netflix.zuul.FilterLoader;
import com.netflix.zuul.FilterUsageNotifier;
import com.netflix.zuul.RequestCompleteHandler;
import com.netflix.zuul.context.SessionContextDecorator;
import com.netflix.zuul.netty.server.BaseServerStartup;
import com.netflix.zuul.netty.server.DirectMemoryMonitor;
import com.netflix.zuul.netty.server.NamedSocketAddress;
import com.netflix.zuul.netty.server.SocketAddressProperty;
import com.netflix.zuul.netty.server.ZuulServerChannelInitializer;
import com.netflix.zuul.netty.server.http2.Http2SslChannelInitializer;
import com.netflix.zuul.netty.ssl.BaseSslContextFactory;

import io.netty.channel.ChannelInitializer;
import io.netty.channel.ChannelOption;
import io.netty.channel.group.ChannelGroup;
import io.netty.handler.ssl.ClientAuth;

/**
 * ApiGw Server Startup - class that configures the Netty server startup settings
 */
@Singleton
public class ApiGwServerStartup extends BaseServerStartup {
    private static final Logger log = LoggerFactory.getLogger(ApiGwFiltersModule.class);

    enum ServerType {
        HTTP,
        HTTP2,
        // HTTP_MUTUAL_TLS,
    }

    private static final String[] WWW_PROTOCOLS = new String[]{"TLSv1.3", "TLSv1.2", "TLSv1.1", "TLSv1"};
    private static final ServerType SERVER_TYPE = ServerType.HTTP2;

    @Inject
    public ApiGwServerStartup(ServerStatusManager serverStatusManager, FilterLoader filterLoader,
                            SessionContextDecorator sessionCtxDecorator, FilterUsageNotifier usageNotifier,
                            RequestCompleteHandler reqCompleteHandler, Registry registry,
                            DirectMemoryMonitor directMemoryMonitor, EventLoopGroupMetrics eventLoopGroupMetrics,
                            EurekaClient discoveryClient, ApplicationInfoManager applicationInfoManager,
                            AccessLogPublisher accessLogPublisher) {
        super(serverStatusManager, filterLoader, sessionCtxDecorator, usageNotifier, reqCompleteHandler, registry,
                directMemoryMonitor, eventLoopGroupMetrics, discoveryClient, applicationInfoManager, accessLogPublisher);
        log.info("API-GATEWAY:Startup: Initialization...");

        // Call PostConstruct by hand, cause inject.*.PostConstruct annotation issue...
        PostConstruct();
    }

    public void PostConstruct() {
        log.debug("API-DATEWAY:Startup: PostConstruct");
        try {
            init();
        } catch (Exception e) {
            log.error("Failed to server netty server. Quitting.");
        }
    }

    @Override
    protected Map<NamedSocketAddress, ChannelInitializer<?>> chooseAddrsAndChannels(ChannelGroup clientChannels) {
        Map<NamedSocketAddress, ChannelInitializer<?>> addrsToChannels = new HashMap<>();

        try {
            int port = new DynamicIntProperty("zuul.server.port", 8000).get();
            String addr = new DynamicStringProperty("zuul.server.addr", "127.0.0.1").get();
            SocketAddress sockAddr = new SocketAddressProperty(addr, "=" + port).getValue();
            String metricId = sockAddr.toString();
            if (sockAddr instanceof InetSocketAddress) {
                metricId = String.valueOf(((InetSocketAddress) sockAddr).getPort());
            }
            // log.debug("API-GATEWAY:chooseAddrsAndChannels:$metricId: " +metricId);

            String mainListenAddressName = "main";
            ServerSslConfig sslConfig;
            ChannelConfig channelConfig = defaultChannelConfig(mainListenAddressName);
            ChannelConfig channelDependencies = defaultChannelDependencies(mainListenAddressName);

            /* These settings may need to be tweaked depending if you're running behind an ELB HTTP listener, TCP listener,
            * or directly on the internet.
            */
            switch (SERVER_TYPE) {
                /* The below settings can be used when running behind an ELB HTTP listener that terminates SSL for you
                * and passes XFF headers.
                */
                case HTTP:
                    channelConfig.set(CommonChannelConfigKeys.allowProxyHeadersWhen, StripUntrustedProxyHeadersHandler.AllowWhen.ALWAYS);
                    channelConfig.set(CommonChannelConfigKeys.preferProxyProtocolForClientIp, false);
                    channelConfig.set(CommonChannelConfigKeys.isSSlFromIntermediary, false);
                    channelConfig.set(CommonChannelConfigKeys.withProxyProtocol, false);

                    logAddrConfigured(sockAddr);
                    addrsToChannels.put(
                        new NamedSocketAddress("http", sockAddr),
                        new ZuulServerChannelInitializer(metricId, channelConfig, channelDependencies, clientChannels)
                    );
                    break;

                /* The below settings can be used when running behind an ELB TCP listener with proxy protocol, terminating
                * SSL in Zuul.
                */
                case HTTP2:
                    // sslConfig = new ServerSslConfig(
                    //     WWW_PROTOCOLS,
                    //     ServerSslConfig.getDefaultCiphers(),
                    //     loadFromResources("ssl/api-gateway-svc.cert"),
                    //     loadFromResources("ssl/api-gateway-svc-pem.key"),
                    //     ClientAuth.NONE
                    // );
                    sslConfig = ServerSslConfig.withDefaultCiphers(
                            loadFromResources("ssl/api-gateway-svc.cert"),
                            loadFromResources("ssl/api-gateway-svc.pem"),
                            WWW_PROTOCOLS
                    );

                    channelConfig.set(CommonChannelConfigKeys.allowProxyHeadersWhen, StripUntrustedProxyHeadersHandler.AllowWhen.NEVER);
                    channelConfig.set(CommonChannelConfigKeys.preferProxyProtocolForClientIp, true);
                    channelConfig.set(CommonChannelConfigKeys.withProxyProtocol, true);
                    channelConfig.set(CommonChannelConfigKeys.isSSlFromIntermediary, false);
                    channelConfig.set(CommonChannelConfigKeys.serverSslConfig, sslConfig);
                    channelConfig.set(CommonChannelConfigKeys.sslContextFactory, new BaseSslContextFactory(registry, sslConfig));

                    // channelConfig.set(ChannelOption.SO_KEEPALIVE, true);
                    // channelConfig.set(ChannelOption.TCP_NODELAY, true);
                    // channelConfig.set(ChannelOption., );

                    addHttp2DefaultConfig(channelConfig, String.valueOf(port));

                    logAddrConfigured(sockAddr, sslConfig);
                    addrsToChannels.put(new NamedSocketAddress("https", sockAddr), new Http2SslChannelInitializer(metricId, channelConfig, channelDependencies, clientChannels));
                    break;

                /* The below settings can be used when running behind an ELB TCP listener with proxy protocol, terminating
                * SSL in Zuul.
                *
                * Can be tested using certs in resources directory:
                *  curl https://localhost:7001/test -vk --cert src/main/resources/ssl/client.cert:zuul123 --key src/main/resources/ssl/client.key
                */
                // case HTTP_MUTUAL_TLS:
                //     sslConfig = new ServerSslConfig(
                //             WWW_PROTOCOLS,
                //             ServerSslConfig.getDefaultCiphers(),
                //             loadFromResources("server.cert"),
                //             loadFromResources("server.key"),
                //             // null,
                //             ClientAuth.REQUIRE,
                //             loadFromResources("truststore.jks"),
                //             loadFromResources("truststore.key"),
                //             false);

                //     channelConfig.set(CommonChannelConfigKeys.allowProxyHeadersWhen, StripUntrustedProxyHeadersHandler.AllowWhen.NEVER);
                //     channelConfig.set(CommonChannelConfigKeys.preferProxyProtocolForClientIp, true);
                //     channelConfig.set(CommonChannelConfigKeys.isSSlFromIntermediary, false);
                //     channelConfig.set(CommonChannelConfigKeys.withProxyProtocol, true);
                //     channelConfig.set(CommonChannelConfigKeys.serverSslConfig, sslConfig);
                //     channelConfig.set(CommonChannelConfigKeys.sslContextFactory, new BaseSslContextFactory(registry, sslConfig));

                //     addrsToChannels.put(new NamedSocketAddress("http_mtls", sockAddr), new Http1MutualSslChannelInitializer(metricId, channelConfig, channelDependencies, clientChannels));
                //     logAddrConfigured(sockAddr, sslConfig);
                //     break;
            }

            log.debug("$addrsToChannel:", addrsToChannels);
            // Class s = server.getClass();
            // for (Method method : s.getDeclaredMethods()) {
            //     log.debug(method.getName());
            // }
        } catch (Exception e) {
            log.error("Startup::Exception: " + e.getMessage());
            for (StackTraceElement tr : e.getStackTrace()) {
                log.error(tr.toString());
            }
        }
        
        return addrsToChannels;
    }

    // public static void addHttp2DefaultConfig(ChannelConfig config, String listenAddressName) {
    //     config.add(new ChannelConfigValue<>(CommonChannelConfigKeys.maxConcurrentStreams,
    //             chooseIntChannelProperty(
    //                     listenAddressName,
    //                     "http2.max.concurrent.streams",
    //                     CommonChannelConfigKeys.maxConcurrentStreams.defaultValue())));
    //     config.add(new ChannelConfigValue<>(CommonChannelConfigKeys.initialWindowSize,
    //             chooseIntChannelProperty(
    //                     listenAddressName,
    //                     "http2.initialwindowsize",
    //                     CommonChannelConfigKeys.initialWindowSize.defaultValue())));
    //     config.add(new ChannelConfigValue<>(CommonChannelConfigKeys.maxHttp2HeaderTableSize,
    //             chooseIntChannelProperty(listenAddressName, "http2.maxheadertablesize", 65536)));
    //     config.add(new ChannelConfigValue<>(CommonChannelConfigKeys.maxHttp2HeaderListSize,
    //             chooseIntChannelProperty(listenAddressName, "http2.maxheaderlistsize", 32768)));

    //     // Override this to a lower value, as we'll be using ELB TCP listeners for h2, and therefore the connection
    //     // is direct from each device rather than shared in an ELB pool.
    //     config.add(new ChannelConfigValue<>(CommonChannelConfigKeys.maxRequestsPerConnection,
    //             chooseIntChannelProperty(listenAddressName, "connection.max.requests", 4000)));

    //     config.add(new ChannelConfigValue<>(CommonChannelConfigKeys.http2AllowGracefulDelayed,
    //             chooseBooleanChannelProperty(listenAddressName, "connection.close.graceful.delayed.allow", true)));
    //     config.add(new ChannelConfigValue<>(CommonChannelConfigKeys.http2SwallowUnknownExceptionsOnConnClose,
    //             chooseBooleanChannelProperty(listenAddressName, "connection.close.swallow.unknown.exceptions", false)));
    // }

    private File loadFromResources(String s) {
        return new File(ClassLoader.getSystemResource(s).getFile());
    }
}
